#include "KinshipEmp.h"
#include "MathMatrix.h"
#include "PedigreeAlleleFreq.h"
#include "OutputKin.h"
#include "PreMeta.h"
#include <string>
#include <cstring>
#include <iostream>
#include <stdlib.h>
#include <vector>
#include "math.h"
#include <algorithm>

double KinshipEmp::q = 0.05;
double KinshipEmp::miss = 0.05;

void KinshipEmp::Setup(Pedigree & ped)
{
    printf("Calculating empirical kinship matrix...\n");
    //How many markers we have from the pedfile?
    int count = ped.count;
    allPairs.Dimension(count, count);
    allPairs.Zero();

    Vector X;
    X.Dimension(count);

    //Now decide which markers to use, MAF > 0.05, Miss genotypes rate < 0.05

    //Use only markers with 2 allels
    //outByAlleleCount.Clear();
    //for(int m=0;m<ped.markerCount;m++)
    //	if(ped.CountAlleles(m) != 2) outByAlleleCount.Push(m);	

    //loop through markers to filter by maf
    int N=0;
    for(int i=0;i<ped.markerCount;i++) 
    {
       int alleleCount = CountAlleles(ped, i);
       //if no one is genotyped, mornomorphic, or marker is not diallelic
       if(alleleCount<=1 || alleleCount>2)
	  continue;

       //calculate allele frequencies
       bool tmp = EstimateFrequencies(ped,i,FREQ_ALL); 
       Vector freq	= ped.GetMarkerInfo(i)->freq;
       double max = freq.Max();
       double maf = 0.0;
       if(freq[1]<freq[2]) maf = freq[1];
       else maf = freq[2];
       int n =0;
       for(int p=0;p<count;p++) {
	  if(!ped[p].isGenotyped(i)) n++;
       }
       //calculate the missing rate
       double rate=n/count;
       if(freq.Length()==3 && max !=1 && maf>q && rate < miss) 
       {
	  N++;
	  double mean, variance;
	  int idx =0;
	  MarkerInfo * info = ped.GetMarkerInfo(i);
	  info->IndexAlleles();
	  String one = info->alleleLabels[1];
	  String two = info->alleleLabels[2];

	  //now fill in the genotypes
	  for(int idv=0;idv<ped.count;idv++)
	  {
	     if(ped[idv].isGenotyped(i)) 
	     {
		Alleles geno = ped[idv].markers[i];
		String genoOne = info->GetAlleleLabel(geno.one);
		String genoTwo = info->GetAlleleLabel(geno.two);
		if(genoOne != genoTwo) X[idx]=1;
		else if(genoOne==one && genoTwo==one) X[idx]=0;
		else if(genoOne==two && genoTwo==two) X[idx]=2;
		else printf("error, more alleles than saved in alleleLables!\n");
		idx++;
	     } 
	     else 
	     { 
		X[idx] = 2*maf; 
		idx++; 
	     }
	  }

	  X.AveVar(mean,variance);	
	  //for(int i=0;i<count;i++) printf("%g ",X[i]);
	  //Now need to calculate the correlation term for marker m
	  X.Add(-1.0*mean);
	  for(int r=0;r<count;r++)
	  {
	     for(int c=r;c<count;c++)
	     {
		allPairs[r][c] += X[r]*X[c]/variance;
	     }
	  }
       }
    }

    for(int i=0;i<count;i++)
       for(int j=i+1;j<count;j++)
	  allPairs[j][i] = allPairs[i][j];

    allPairs.Multiply(1.0/N);
    //for(int row=0;row<5;row++) 
    //{
    //	for(int col=0;col<5;col++)
    //	{ 
    //		printf("%g ",allPairs[row][col]);
    //	}
    //	printf("\n");
    //}
    printf("Finished calculating empirical kinship matrix. \n ");
    if(OutputKin::outputKin) 
    {
       FILE * file = NULL;
       String filename;
       if(PreMeta::outputFile != "")
       {
	  filename.printf("Empirical.Kinship.%s",(const char *) PreMeta::outputFile);
       }
       else
       {
	  filename.printf("Empirical.Kinship%s",(const char *) PreMeta::outputFile);
       }
       file = fopen(filename, "wt");
       for(int r=0;r<allPairs.rows;r++)
       {
	  for(int c=0;c<allPairs.cols;c++)
	  {
	     fprintf(file,"%g ",allPairs[r][c]);
	  }
	  fprintf(file,"\n");
       }
       fclose(file);
    }
}

void KinshipEmp::SetupVCF(Pedigree & ped, StringIntHash & foundersHash)
{
   printf("Calculating empirical kinship matrix...\n");
   VcfFileReader reader;
   VcfHeader header;
   VcfRecord record;
   reader.open(PreMeta::vcfInput,header);

   int numSamples = header.getNumSamples();

   allPairs.Dimension(numSamples,numSamples);
   allPairs.Zero();

   Vector genotype;
   genotype.Dimension(numSamples);
   int N=0;
   if(PreMeta::dosage)
   {
      while(reader.readRecord(record))
      {
	 //step1: decide if this marker should be included
	 int n=0,nf=0,nmiss=0;
	 double fmaf=0.0,maf=0.0;
	 VcfRecordGenotype & genoInfo = record.getGenotypeInfo();
	 for(int s=0;s<numSamples;s++)
	 {
	    const std::string * geno = genoInfo.getString("DS",s);
	    const char * sample = header.getSampleName(s);

	    if(*geno == ".")
	    {
	       nmiss++;
	       continue;
	    }
	    float dose = atof((*geno).c_str());

	    //judge if this is a founder
	    int founder = foundersHash.Integer(sample);
	    if(founder != -1)
	    {
	       fmaf += dose;
	       nf++;
	    }
	    //Calculate overall maf
	    maf += dose;
	    n++;
	 }

	 //calculate final maf
	 if(nf!=0)
	    fmaf /= 2.0*nf;
	 maf /= 2.0*n;
	 if(fmaf==0.0 && maf >0.0)
	    fmaf=maf;

	 if(PreMeta::FounderFreq)
	 {
	    if(nmiss > numSamples*miss || fmaf<q)
	    {
	       continue;
	    }
	 }
	 else
	 {
	    if(nmiss > numSamples*miss || maf<q)
	    {
	       continue;
	    }
	 }
	 N++;

	 //step2: if yes, then loop through all samples to get genotype. 
	 //genotype.Clear();
	 double mean = 0.0, mean_sqr= 0.0;
	 int n_ = 0;
	 for(int s =0;s<numSamples;s++)
	 {
	    //fill in genotype vector for this sample
	    const std::string * geno = genoInfo.getString("DS",s);
	    if(*geno == ".")
	    {
	       genotype[s] = _NAN_;
	    }
	    else
	    {
	       genotype[s] = atof((*geno).c_str());
	       mean += genotype[s];
	       n_++;
	       mean_sqr += genotype[s]*genotype[s];
	    }
	 }

	 //step 3: standardize genotype vector
	 mean /= n_;
	 double variance = mean_sqr/n_ - mean*mean;
	 for(int i=0;i<genotype.Length();i++)
	 {
	    if(genotype[i]==_NAN_)
	    {
	       //make sure if genotype is missing, then it is not contributing
	       genotype[i]=0.0;
	    }
	    else
	    {
	       genotype[i] -= mean;
	    }
	 }
	 //add up to kinship
	 for(int r=0;r<numSamples;r++)
	 {
	    for(int c=r;c<numSamples;c++)
	    {
	       allPairs[r][c] += genotype[r]*genotype[c]/variance;
	    }
	 }
      }

      for(int i=0;i<numSamples;i++)
	 for(int j=i+1;j<numSamples;j++)
	    allPairs[j][i] = allPairs[i][j];

      allPairs.Multiply(1.0/N);
   }

   if(PreMeta::genoFromVCF)
   {
      while(reader.readRecord(record))
      {
	 //step1: decide if this marker should be included
	 double fmaf=0.0,maf=0.0;
	 int nmiss =0;
	 std::vector<int> fcounts;
	 std::vector<int> counts;

	 for(int i = 0; i < numSamples; i++)
	 {
	    const char * sample = header.getSampleName(i);
	    int numGTs = record.getNumGTs(i);
	    Vector allele;
	    for(int j = 0; j < numGTs; j++)
	    {
	       allele.Push(record.getGT(i,j));
	    }
	    int bad =0;
	    for(int m=0;m<allele.Length();m++)
	    {
	       bad += (allele[m]==VcfGenotypeSample::MISSING_GT);
	    }
	    if(bad>0)
	    {
	       //there is at least one allele missing, so skip this sample
	       nmiss++;
	       continue;
	    }

	    //judge if this is a founder
	    int founder = foundersHash.Integer(sample);

	    for(int m=0;m<allele.Length();m++)
	    {
	       if(allele[m] >= counts.size())
	       {
		  counts.resize(allele[m]+1);
	       }
	       ++counts[allele[m]];
	       if(founder != -1)
	       {
		  if(allele[m] >= fcounts.size())
		  {
		     fcounts.resize(allele[m]+1);
		  }
		  ++fcounts[allele[m]];
	       }
	    }
	    //Only consider diallelic markers
	    if(counts.size()>2)
	    {
	       break;
	    }
	 }

	 if(counts.size()>2 || counts.size()==1 || fcounts.size()==1)
	 {
	    continue;
	 }

	 //Calculate maf
	 double sum = 0.0;
	 for(int i=0;i<counts.size();i++)
	 {
	    sum+= counts[i];
	 }
	 maf = counts[1]/sum;
	 sum = 0.0;
	 for(int i=0;i<fcounts.size();i++)
	 {
	    sum+= fcounts[i];
	 }
	 if(sum>0)
	    fmaf = fcounts[1]/sum;
	 if(fmaf==0.0 && maf >0.0)
	    fmaf=maf;

	 if(PreMeta::FounderFreq)
	 {
	    if(nmiss > numSamples*miss || fmaf<q)
	    {
	       continue;
	    }
	 }
	 else
	 {
	    if(nmiss > numSamples*miss || maf<q)
	    {
	       continue;
	    }
	 }

	 N++;
	 //printf("%d\n",N);

	 //step2: if yes, then loop through all samples to get genotype. 

	 int n=0;
	 double mean = 0.0,mean_sqr=0.0;
	 for(int s=0;s<numSamples;s++)
	 {
	    int numGTs = record.getNumGTs(s);
	    Vector allele;

	    for(int j = 0; j < numGTs; j++)
	    {
	       allele.Push(record.getGT(s,j));
	    }
	    int bad =0;
	    for(int m=0;m<allele.Length();m++)
	    {
	       bad += (allele[m]==VcfGenotypeSample::MISSING_GT);
	    }
	    //if genotype is missing, set genotype to be zero for no contribution
	    if(bad>0)
	    {
	       genotype[s] = _NAN_;
	       continue;
	    }
	    int sum=0;
	    for(int i=0;i<allele.Length();i++)
	       sum+=allele[i];
	    genotype[s] = sum;
	    mean += sum;
	    n++;
	    mean_sqr += sum*sum;
	 }

	 //step 3: standardize genotype vector
	 mean /= n;
	 double variance = mean_sqr/n - mean*mean;
	 //printf("%d %g %g\n",N,mean,variance);
	 for(int i=0;i<genotype.Length();i++)
	 {
	    if(genotype[i]==_NAN_)
	    {
	       //make sure if genotype is missing, then it is not contributing
	       genotype[i]=0.0;
	    }
	    else
	    {
	       genotype[i] -= mean;
	    }
	 }
	 //add up to kinship
	 for(int r=0;r<numSamples;r++)
	 {
	    for(int c=r;c<numSamples;c++)
	    {
	       allPairs[r][c] += genotype[r]*genotype[c]/variance;
	    }
	 }
      }

      for(int i=0;i<numSamples;i++)
	 for(int j=i+1;j<numSamples;j++)
	    allPairs[j][i] = allPairs[i][j];

      allPairs.Multiply(1.0/N);
   }

   if(OutputKin::outputKin)
   {
      FILE * file = NULL;
      String filename;
       if(PreMeta::outputFile != "")
       {
          filename.printf("Empirical.Kinship.%s",(const char *) PreMeta::outputFile);
       }  
       else
       {
          filename.printf("Empirical.Kinship%s",(const char *) PreMeta::outputFile);     
       }
      file = fopen(filename, "wt");
      for(int r=0;r<allPairs.rows;r++)
      {
	 for(int c=0;c<allPairs.cols;c++)
	 {
	    fprintf(file,"%g ",allPairs[r][c]);

	 }
	 fprintf(file,"\n");
      }
      fclose(file);
   }
}


